"""
This is a module for working with constants, including storing values for different units and finding conversion factors.

"""

prefix_exps_dict = {'y': -24, 'z': -21, 'a': -18, 'f': -15,
                    'p': -12, 'n': -9, 'mu': -6, 'm': -3,
                    'c': -2, 'd': -1, 'da': 1, 'h': 2,
                    'k': 3, 'M': 6, 'G': 9, 'T': 12,
                    'P': 15, 'E': 18, 'Z': 21, 'Y': 24}

energy_units_names = {'eV': ['ev', 'electron-volt', 'electronvolt', 'electronvolts', 'electron-volts'],
                      'Ha': ['ha', 'hartree', 'eh', 'e_h', 'hartrees'],
                      'J': ['j', 'joule', 'joules'],
                      'Ry': ['ry', 'rydberg', 'rydbergs']}

energy_units_vals = {'eV': (2.7211396132, 1), 'Ha': (1, 0), 'J': (4.359748199, -18), 'Ry': (0.5, 0)}

length_units_names = {'bohr': ['bohr', 'a.u', 'atomic units', 'au'], 'angstrom': ['angstrom, a'], 'm': ['m', 'meter']}

length_units_vals = {'bohr': (1, 0), 'angstrom': (0.529177249, 0), 'm': (5.29177249, -11)}

recip_points_units_names = {'cartesian': ['tpiba', 'cartesian', 'cart'], 'crystal': ['crystal', 'cryst', 'frac', 'fractional']}


def prefix_exp(prefix):
    """"
    Finds the exponent corresponding to a prefix.
    For example, the prefix 'c' (centi) corresponds to 1e-2, so the exponent returned would be -2.

    Parameters
    ----------
    prefix : str
        The 1-2 letter case-sensitive prefix.

    Returns
    -------
    exponent : int
        The exponent corresponding to the prefix.

    Raises
    ------
    ValueError
        If `prefix` is not in prefix_exps_dict.keys(). These are:
            ['y', 'z', 'a', 'f', 'p', 'n', 'mu', 'm', 'c', 'd',
             'da', 'h', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']

    Examples
    --------
    >>> prefix_exp('c')
    -2
    >>> prefix_exp('mu')
    -6
    >>> prefix_exp(k)
    3

    """

    if prefix == '':
        return 0
    else:
        if prefix not in prefix_exps_dict.keys():
            raise ValueError(f"Please choose prefixes from the following list: {list(prefix_exps_dict.keys())}")

        return prefix_exps_dict[prefix]


def find_prefix_and_base_units(user_input_units, units_dict):
    """
    Dissects units to a prefix and base units, both with standardized names as specified by units_dict.
    For example, for the units 'meV', the prefix is 'm' and the base units are 'eV'.

    Parameters
    ----------
    user_input_units : str
        The units to be analyzed for its prefix and base.

    units_dict : dict
        A dictionary specifying the standard unit name corresponding to a set of possible units names (case-insensitive).

        For example, several possible variations of the unit hatree include Ha, hartree, etc. If Ha is
        chosen as the standard name, then the dictionary entry will be:
        {'Ha': ['ha', 'hartree', 'eh', 'e_h', 'hartrees']}

        Adding the unit Joule to the dict, and setting its standard name to 'J':
        {'Ha': ['ha', 'hartree', 'eh', 'e_h', 'hartrees'], 'J': ['j', 'joule', 'joules']}

    Returns
    -------
    prefix : str
        The standardized prefix.

    standard_units_name : str
        The standardized base unit name.

    Raises
    ------
    ValueError
        If the user_input_units, after removing any prefixes, is not in the keys or values of the units_dict.

    Examples
    --------
    >>> find_prefix_and_base_units('nm', {'m':['m', 'meter']})
    ('n', 'm')
    >>> find_prefix_and_base_units('nmeter', {'m':['m', 'meter']})
    ('n', 'm')
    >>> find_prefix_and_base_units('meter', {'m':['m', 'meter']})
    ('', 'm')
    >>> find_prefix_and_base_units('cm', {'m':['m', 'meter']}, 'bohr': ['bohr', 'a.u'])
    ('c', 'm')
    >>> find_prefix_and_base_units('a.u.', {'m':['m', 'meter']}, 'bohr': ['bohr', 'a.u.'])
    ('', 'bohr')


    """

    user_input_units = user_input_units.replace(' ', '').replace('-', '')

    # only case where prefix may get confused
    if user_input_units == 'mm':
        return 'm', 'm'

    for units_name in units_dict.keys():
        for name_variation in units_dict[units_name]:
            if name_variation in user_input_units.lower():
                user_input_units_split = user_input_units.lower().split(name_variation)

                if len(user_input_units_split) == 2 and user_input_units_split[1] == '':

                    prefix = user_input_units[:len(user_input_units_split[0])]
                    standard_units_name = units_name

                    return prefix, standard_units_name

    raise ValueError(f"Please choose units from the following list: {list(units_dict.keys())}")


def standardize_units_name(user_input_units, units_dict):
    """
    Converts any units to their standard name as specified in units_dict. Uses find_prefix_and_base_units to convert
    user_input_units to a prefix and standardized base units, then combines them into one string.

    Parameters
    ----------
    user_input_units : str
        The units to be standardized

    units_dict : dict
        A dictionary specifying the standard unit name corresponding to a set of possible units names (case-insensitive).
        See prefix_and_base_units for details.

    Returns
    -------
    standard_units : str
       The standardized units name.

    Examples
    --------
    >>> find_prefix_and_base_units('nm', {'m':['m', 'meter']})
    'nm'
    >>> find_prefix_and_base_units('nmeter', {'m':['m', 'meter']})
    'nm'
    >>> find_prefix_and_base_units('a.u.', {'m':['m', 'meter']}, 'bohr': ['bohr', 'a.u.'])
    'bohr'

    """
    prefix, units = find_prefix_and_base_units(user_input_units, units_dict)

    return prefix + units


def conversion_factor(init_units, final_units, units_names, units_vals):
    """
    Finds the conversion factor between two units.

    Parameters
    ----------
    init_units : str
        The initial units in the conversion.

    final_units : str
        The final units in the conversion.

    units_names : dict
        A dictionary specifying the standard unit name corresponding to a set of possible units names (case-insensitive).
        See prefix_and_base_units for details.

    units_vals : dict
        A dictionary specifying the conversion factors between different units, with units labeled by their standard name as
        specified in units_names. Values are represented as tuples (base, exponent).

        Example: For energies, possible units include hartrees and electron-volts.' Setting one value (hartree = 1), the remaining values
                 are the conversion factors between different units and hartrees.

                 energy_units_vals = {'eV': (2.7211396132, 1), 'Ha': (1, 0),'Ry': (0.5, 0)}

    Returns
    -------
    conversion_factor : float
       The conversion factor to convert from init_units to final_units.

    Raises
    ------
    ValueError
        If the base units of `init_units` or `final_units` are not in the keys of `units_vals`

    Examples
    --------
        >>> conversion_factor('fm', 'cm', {m': ['m', 'meter']},
                                          {'m': (5.29177249, -11)})
        1e-13
        >>> conversion_factor('a.u', 'bohr', {'bohr': ['bohr', 'a.u'], 'angstrom': ['angstrom, a']},
                                             {'bohr': (1, 0), 'angstrom': (0.529177249, 0)})
        1.0
        >>> conversion_factor('a.u', 'angstrom', {'bohr': ['bohr', 'a.u'], 'angstrom': ['angstrom, a']},
                                                 {'bohr': (1, 0), 'angstrom': (0.529177249, 0)})
        0.529177249
         
    """

    init_prefix, init_units = find_prefix_and_base_units(init_units, units_names)
    final_prefix, final_units = find_prefix_and_base_units(final_units, units_names)

    if init_units not in units_vals.keys() or final_units not in units_vals.keys():
        raise ValueError(f"Base units {init_units}, {final_units} are not in the units_vals dictionary: {list(units_vals.keys())}")

    init_val = units_vals[init_units]
    final_val = units_vals[final_units]

    conversion_factor = (final_val[0] / init_val[0],
                         (final_val[1] - prefix_exp(final_prefix)) - (init_val[1] - prefix_exp(init_prefix)))

    return conversion_factor[0] * 10**(conversion_factor[1])


def energy_conversion_factor(init_units, final_units):
    """
    find the conversion factor between two energy units.

    Parameters
    ----------
    init_units : str
        The initial units in the conversion.

    final_units : str
        The final units in the conversion.

    Returns
    -------
    conversion_factor : float
        The conversion factor to convert from init_units to final_units.

    Examples
    --------
    >>> energy_conversion_factor('meV', 'Ha')
    3.674930882447527e-05
    >>> energy_conversion_factor('Ry', 'Ha')
    2.0
       
    """

    return conversion_factor(init_units, final_units, energy_units_names, energy_units_vals)


def length_conversion_factor(init_units, final_units):
    """
    find the conversion factor between two length units.

    Parameters
    ----------
    init_units : str
        The initial units in the conversion.

    final_units : str
        The final units in the conversion.

    Returns
    -------
    conversion_factor : float
        The conversion factor to convert from init_units to final_units.

    Examples
    --------
    >>> length_conversion_factor('a.u', 'bohr')
    1.0
    >>> length_conversion_factor('a.u', 'nm')
    5.29177249e-2

    """

    return conversion_factor(init_units, final_units, length_units_names, length_units_vals)


def hbar(units):
    """
    find the value of hbar for specific units.

    Parameters
    ----------
    units : str
       The units hbar should be returned in.

    Returns
    -------
    hbar : float
       The value of hbar in the specified units.

    Raises
    ------
    ValueError
        If `units` is not in the keys of `hbar_dict`

    """
    hbar_dict = {'ev*fs': (6.582119569, -1), 'ev*s': (6.582119569, -16), 'atomic': (1, 0), 'J*s': (1.054571817, -34)}

    if units not in hbar_dict.keys():
        raise ValueError(f"Please choose hbar units from the following list: {list(hbar_dict.keys())}")

    return hbar_dict[units][0] * (10 ** hbar_dict[units][1])



import os
import numpy as np
from perturbopy.io_utils.io import open_yaml


class CalcMode():
    """
    This is a general class representation of a Perturbo calculation mode.

    Parameters
    ----------
    _pert_dict : dict
       dictionary containing inputs and outputs from a Perturbo calculation, as well as
       basic data of the material. Typically obtained from the YAML file outputted by Perturbo.
    alat : float
    alat_units : str
    lat : array
    lat_units : str
    recip_lat : array
    recip_lat_units : str
    nat : int
    atomic_pos : array
    atomic_pos_units : str
    volume : float
    volume_units : str
    nsym : int
    kc_dim : array
    polar_alpha : float
    epsil : array
    qc_dim : array
    mass : array
    mass_units : str
    symop : array
    zstart : array
    system_2d : bool
    num_wann : int
    wannier_center : array
    wannier_center_cryst : array

    """

    def __init__(self, pert_dict):
        """
        Constructor method

        """

        # Extract calculation mode name from pert_dict
        self.calc_mode = pert_dict['input parameters']['after conversion'].pop('calc_mode')

        # Extract basic data from pert_dict
        self.alat = pert_dict['basic data']['alat']
        self.alat_units = pert_dict['basic data']['alat units']
        self.lat = np.transpose(np.array(pert_dict['basic data']['lattice vectors']))
        self.lat_units = pert_dict['basic data']['lattice vectors units']
        self.recip_lat = np.transpose(np.array(pert_dict['basic data']['reciprocal lattice vectors']))
        self.recip_lat_units = pert_dict['basic data']['reciprocal lattice vectors units']
        self.nat = pert_dict['basic data']['number of atoms in unit cell']
        self.atomic_pos = np.transpose(np.array(pert_dict['basic data']['atomic positions']))
        self.atomic_pos_units = pert_dict['basic data']['atomic positions units']
        self.volume = pert_dict['basic data']['volume']
        self.volume_units = pert_dict['basic data']['volume units']
        self.nsym = pert_dict['basic data']['number of symmetry operations']
        self.kc_dim = np.array(pert_dict['basic data']['kc dimensions'])
        self.polar_alpha = pert_dict['basic data']['polar_alpha']
        self.epsil = np.transpose(np.array(pert_dict['basic data']['epsil']))
        self.qc_dim = np.array(pert_dict['basic data']['qc dimensions'])
        self.mass = np.array(pert_dict['basic data']['mass'])
        self.mass_units = pert_dict['basic data']['mass units']
        self.symop = [pert_dict['basic data']['symop']]
        self.zstar = [pert_dict['basic data']['zstar']]
        self.system_2d = pert_dict['basic data']['system_2d']
        self.num_wann = pert_dict['basic data']['number of Wannier functions']
        self.wannier_center = np.transpose(np.array(pert_dict['basic data']['wannier_center']))
        self.wannier_center_cryst = np.transpose(np.array(pert_dict['basic data']['wannier_center_cryst']))

        # Store remaining, unprocessed data from pert_dict (often, such data is calculation mode specific)
        self._pert_dict = pert_dict

    @classmethod
    def from_yaml(cls, yaml_path='pert_output.yml'):
        """
        Class method to create a CalcMode object from the YAML file
        generated by a Perturbo calculation.

        Parameters
        ----------
        yaml_path : str, optional
           Path to the YAML file generated by a Perturbo calculation

        Returns
        -------
        calc_mode : CalcMode
           The CalcMode object generated from the YAML file

        """

        if not os.path.isfile(yaml_path):
            raise FileNotFoundError(f'File {yaml_path} not found')

        yaml_dict = open_yaml(yaml_path)

        return cls(yaml_dict)



"""
   Utils to select which tests to run based on the command line arguments and
   test tags.
"""
import os
import copy
import shutil
import sys
from perturbopy.io_utils.io import open_yaml


def read_test_tags(test_name, func_name, source_folder):
    """
    Get a list of tags for a given test. List of tags is combined from the tags from
    pert_input.yml and epr_info.yml for a given epr file.

    Parameters
    ----------
    test_name : str
        name of the folder inside the tests/ folder

    func_name : str
        name of the test programm, which we run
        (do we test perturbo or qe2pert)
    source_folder : str
        name of the folder, where should be all the testing supplementary files (reference, input files, etc.)
    Returns
    -------
    tag_list : list
        list of tags for a given test
    epr_name : str
        name of the epr file associated with this test
    """
    
    epr_dict_path = os.path.join(source_folder, 'epr_info.yml')
    epr_info = open_yaml(epr_dict_path)

    if (func_name == 'test_perturbo') or (func_name == 'test_perturbo_for_qe2pert'):
        driver_path_suffix = 'inputs/' + test_name
        perturbo_driver_dir_path = [x[0] for x in os.walk(source_folder) if x[0].endswith(driver_path_suffix)][0]
        pert_input = open_yaml(f'{perturbo_driver_dir_path}/pert_input.yml')

        # Read the tags from pert_input.yml
        input_tags = []
        if 'tags' in pert_input['test info'].keys():
            input_tags = pert_input['test info']['tags']

        # Read the tags from epr_info.yml
        epr_name = pert_input['test info']['epr']

        epr_tags = []
        if 'tags' in epr_info[epr_name].keys():
            epr_tags = epr_info[epr_name]['tags']

        tag_list = input_tags + epr_tags
        tag_list = sorted(list(set(tag_list)))
    
    elif func_name == 'test_qe2pert':
        if 'tags' in epr_info[test_name]:
            tag_list = epr_info[test_name]['tags']
        epr_name = test_name

    return tag_list, epr_name


def get_all_tests(func_name, source_folder):
    """
    Get the names of all test folders based on the epr_info.yml file.

    Parameters
    ----------
    func_name : str
       name of the metafunction, which we parametrize

    Returns
    -------
    test_folder_list : list
       list of all test names
    source_folder : str
        name of the folder, where should be all the testing supplementary files (reference, input files, etc.)
    """
    test_folder_list = []
    dev_test_folder_list = []

    epr_dict_path = os.path.join(source_folder, 'epr_info.yml')
    epr_info = open_yaml(epr_dict_path)

    if (func_name == 'test_perturbo') or (func_name == 'test_perturbo_for_qe2pert'):

        test_list = ['bands', 'phdisp', 'ephmat']
        for epr in epr_info:
            if 'tests' in epr_info[epr].keys():
                if (func_name == 'test_perturbo'):
                    test_list = epr_info[epr]['tests']

                test_folder_list += [f'{epr}-{t}' for t in test_list]

            if 'devel tests' in epr_info[epr].keys():
                dev_test_list = epr_info[epr]['devel tests']

                dev_test_folder_list += [f'{epr}-{t}' for t in dev_test_list]

    elif func_name == 'test_qe2pert':
        test_folder_list = [epr for epr in epr_info]

    return test_folder_list, dev_test_folder_list


def print_test_info(test_name, input_dict, test_type):
    """
    Print information about a test.

    Parameters
    ----------
    test_name : str
        name of the test folder
    input_dict : dict
        dictionary contatining the test info
    test_type : str
        define that type of testing we make - either 'qe2pert' or 'perturbo'
    """

    if test_type == 'perturbo':
        if 'desc' in input_dict['test info']:
            desc = input_dict['test info']['desc']
        else:
            desc = None
    elif test_type == 'qe2pert':
        if 'description' in input_dict[test_name]:
            desc = input_dict[test_name]['description']
        else:
            desc = None
        
    print(f'\n === Test folder === :\n {test_name}')

    if desc is not None:
        print(f'\n === Description === :\n {desc}')

    sys.stdout.flush()


def filter_tests(all_test_list, tags, exclude_tags, epr, test_names, func_name, run_qe2pert, source_folder):
    """
    Return the list of test folders based on command line options

    Parameters
    ----------
    all_test_list : list
       list of all the test folders

    tags : list or None
       list of tags to include

    exclude_tags : list or None
       list of tags to exclude

    epr : list or None
       list of the epr files

    test_names : list or None
       list of test folders to include
    
    func_name : str
        name of the test programm, which we run
        (do we test perturbo or qe2pert)
    
    run_qe2pert : bool
        whether perturbo_for_qe2pert tests are conducted or not

    source_folder : str
        name of the folder, where should be all the testing supplementary files (reference, input files, etc.)

    Returns
    -------
    test_list : list
       list of test for a given pytest run

    Raises
    ------
    RuntimeError
       if --test-names and (--tags or --exclude_tags) are specified at the same time
       or
       if no test folders are selected

    ValueError
       if --test-names contains a name of a test that is not present
    """

    test_list = copy.deepcopy(all_test_list)

    # sort based on tags
    if tags is not None or exclude_tags is not None or epr is not None:
        for test_name in all_test_list:

            # tags for a given test
            test_tag_list, epr_name = read_test_tags(test_name, func_name, source_folder)

            # tags from command line
            if tags is not None:

                keep_test = False

                for tag in tags:
                    if tag in test_tag_list:
                        keep_test = True
                        break

                if not keep_test:
                    test_list.remove(test_name)

            # exclude tags from command line
            if exclude_tags is not None:

                keep_test = True

                for tag in exclude_tags:
                    if tag in test_tag_list:
                        keep_test = False
                        break

                if not keep_test and test_name in test_list:
                    test_list.remove(test_name)

            # epr file name
            if epr is not None:

                if epr_name not in epr and test_name in test_list:
                    test_list.remove(test_name)

    # test name from command line
    if test_names is not None:

        if tags is not None or exclude_tags is not None:
            errmsg = ('If the --test-names option is specified, \n'
                      '--tags and --exclude_tags must NOT be specified.'
                     )

            raise RuntimeError(errmsg)

        for test_name_cmd in test_names:
            if test_name_cmd not in all_test_list:
                if (func_name == 'test_perturbo') or (func_name == 'test_qe2pert'):
                    errmsg = (f'Test {test_name_cmd} is not listed in epr_info.yml, \n'
                              f'but specified in --test-names option. Full test_list: {test_list}'
                             )
                    raise ValueError(errmsg)
                elif (func_name == 'test_perturbo_for_qe2pert') and run_qe2pert:
                    errmsg = (f'Test {test_name_cmd} is not listed for running on the perturbo run for \n'
                              'qe2pert check but specified in --test-names option. On this run, only \n'
                              f'this tests supposed to run: {test_list}'
                             )
                    raise ValueError(errmsg)

        test_list = test_names

    if not test_list:
        raise RuntimeError('No test folders selected')

    print('\n\n === Test folders == :')
    print(' \n'.join(test_list))
    print('')
    sys.stdout.flush()

    return test_list


def setup_default_tol(igns_n_tols, test_case):
    """
    Setup the default tolerances for each file to compare if the tolerances are
    not specified in the pert_input.yml file.

    This function ensures that every output file to compare has the following
    dictionary structure:

    .. code-block :: python

        output_file.yml:
            abs tol:
                default:
                    1e-8

            # relative error
            rel tol:
                default:
                    0.01

    The elements are considerent different if the following equation does not apply:

    >>> absolute(a - b) <= (abs_tol + rel_tol * absolute(b))

    Parameters
    ----------
    igns_n_tols : list
        list of dictionaries, which contain containing the tolerances needed
        to performance comparison of ref_outs and new_outs
    test_case : str
        define what type of the test we run - for perturbo testing or for the
        qe2pert testing.

    Returns
    -------
    igns_n_tols_updated : list
        **updated** list containing the dictionary with tolerances

    """

    if test_case == 'perturbo':
        default_abs_tol = 1e-8
        default_rel_tol = 0.01
    elif test_case == 'perturbo_for_qe2pert':
        # in case of the testing qe2pert relative error could be bigger
        # due to the error that accumulates from scf calculations
        default_abs_tol = 1e-8
        default_rel_tol = 0.01

    igns_n_tols_updated = []

    # run thru all files (their list is the keys set)
    for outfile in igns_n_tols:

        if not isinstance(outfile, dict):
            # if we don't have any information about the errors
            # for this file - define default one
            outfile = {'abs tol':
                       {'default': default_abs_tol},
                       'rel tol':
                       {'default': default_rel_tol}
                      }

        else:
            if test_case == 'perturbo_for_qe2pert':
                # if we test perturbo for qe2pert,
                # we move error for qe2pert testing into the cells
                # `abs_tol` and `rel_tol`
                if 'qe2pert abs tol' in outfile.keys():
                    outfile['abs tol'] = outfile['qe2pert abs tol']
                if 'qe2pert rel tol' in outfile.keys():
                    outfile['rel tol'] = outfile['qe2pert rel tol']

            # if we have some dict but without `abs_tol` key - take default
            if 'abs tol' not in outfile.keys():
                outfile['abs tol'] = {'default': default_abs_tol}

            # if we have some `abs_tol` key, but only for specific cases,
            # add default values
            elif 'default' not in outfile['abs tol'].keys():
                outfile['abs tol']['default'] = default_abs_tol

            # if we have some dict but without `rel_tol` key - take default
            if 'rel tol' not in outfile.keys():
                outfile['rel tol'] = {'default': default_rel_tol}

            # if we have some `rel_tol` key, but only for specific cases,
            # add default values
            elif 'default' not in outfile['rel tol'].keys():
                outfile['rel tol']['default'] = default_rel_tol

        igns_n_tols_updated.append(outfile)

    return igns_n_tols_updated


def get_tol(ig_n_tol, key, exact_match=False):
    """
    Extract the absolute and relative tolerances for ``key`` from ``ig_n_tol`` dict.

    Parameters
    ----------
    ig_n_tol : dict
       dictionary of ignore keywords and tolerances needed to make comparison on values
    key : str
       A key for the tolerance. If this key is not specified in the tolerance dict,
       a default tolerance will be applied.

    exact_match : bool, optional
       if True, the non-default tolerance if applied to a key only if this key
       matches exactly with the one from the ig_n_tol;
       if False, ``str1 in str2`` condition is enough. For example ``phys`` would
       be considered as a key for ``physics`` if exact_match is False.
       Default is False.

    .. note ::
       The ``rel tol`` from the dictionary is assumed to be in **percents**.

    Returns
    -------
    atol : float
       absolute tolerance for key
    rtol : float
       relative tolerance for key
    """

    atol_dict = ig_n_tol['abs tol']

    kname, cond = key_in_dict(key, atol_dict, exact_match=exact_match)

    if cond:
        atol = atol_dict[kname]
    else:
        atol = atol_dict['default']

    rtol_dict = ig_n_tol['rel tol']

    kname, cond = key_in_dict(key, rtol_dict, exact_match=exact_match)

    if cond:
        rtol = rtol_dict[kname]
    else:
        rtol = rtol_dict['default']

    return float(atol), float(rtol) / 100


def key_in_dict(k, d, exact_match=False):
    """
    Check is a key is in dictionary.

    Parameters
    ----------
    k : str
       A dictionary key.
    d : dict
       A dictionary.

    exact_match : bool
       Whether apply or not the exact match condition for the key

    Returns
    -------
    key : str
       A key from dictionary d that contains k

    condition : bool
       True is the key k is in dict d
    """

    if exact_match:
        return k, k in d

    else:

        for kk in d:
            if str(k) in str(kk) or str(kk) in str(k):
                return kk, True
                break

        return k, False
        

def define_nq_num(output_name):
    """
    Define the number of the q-points,
    obtained during the phonon
    calculation

    Parameters:
    output_name : str
        name to the ph output file from which we obtain
        number of q-points

    Returns
    -------
    nq_num : int
       number of q-points
    """
    with open(output_name, "r") as f:
        for i, line in enumerate(f):
            if line.find('q-points)') != -1:
                nq_num = int(line.split()[1])
    return nq_num


def ph_collection(prefix, nq_num):
    """
    Collect the phonon data into a directory called save.
    The save directory contains all the information needed
    for PERTURBO to interface with QE

    Parameters:
    prefix : str
        prefix of this calculation
    nq_num : int
        number of q points in this calculation
    """
    print(os.getcwd())
    
    os.makedirs('save', exist_ok=True)
    
    # remove wfc files in tmp
    if os.path.exists('tmp'):
        for file in os.listdir('tmp'):
            if 'wfc' in file:
                os.remove(f'tmp/{file}')

    dir = 'tmp/_ph0'
    
    shutil.copytree(f'{dir}/{prefix}.phsave', f'save/{prefix}.phsave')
    
    # copy dyn files
    for file in os.listdir('.'):
        if file.startswith(f'{prefix}.dyn'):
            shutil.copy(file, 'save')

    for nq in range(1, nq_num + 1):
        
        # copy dvscf files
        if nq > 1:

            # parallel version
            if os.path.exists(f'{dir}/{prefix}.q_{nq}/{prefix}.dvscf1'):
                shutil.copy(f'{dir}/{prefix}.q_{nq}/{prefix}.dvscf1', f'save/{prefix}.dvscf_q{nq}')
            # serial version
            elif os.path.exists(f'{dir}/{prefix}.q_{nq}/{prefix}.dvscf'):
                shutil.copy(f'{dir}/{prefix}.q_{nq}/{prefix}.dvscf', f'save/{prefix}.dvscf_q{nq}')
            else:
                raise FileNotFoundError(f"{dir}/{prefix}.q_{nq}/{prefix}.dvscf and {dir}/{prefix}.q_{nq}/{prefix}.dvscf1 don't exist")

            for file in os.listdir(f'{dir}/{prefix}.q_{nq}'):
                if file.endswith('wfc'):
                    os.remove(f'{dir}/{prefix}.q_{nq}/{file}')

        else:

            # parallel version
            if os.path.exists(f'{dir}/{prefix}.dvscf1'):
                shutil.copy(f'{dir}/{prefix}.dvscf1', f'save/{prefix}.dvscf_q{nq}')
            # serial version
            elif os.path.exists(f'{dir}/{prefix}.dvscf'):
                shutil.copy(f'{dir}/{prefix}.dvscf', f'save/{prefix}.dvscf_q{nq}')
            else:
                raise FileNotFoundError(f"{dir}/{prefix}.dvscf and {dir}/{prefix}.dvscf1 don't exist")

            for file in os.listdir(dir):
                if 'wfc' in file:
                    os.remove(f'{dir}/{file}')
             
    shutil.copy(f'save/{prefix}.dyn0', f'save/{prefix}.dyn0.xml')



import numpy as np
from perturbopy.postproc.calc_modes.calc_mode import CalcMode
from perturbopy.postproc.dbs.units_dict import UnitsDict
from perturbopy.postproc.dbs.recip_pt_db import RecipPtDB
from perturbopy.postproc.utils.plot_tools import plot_dispersion, plot_recip_pt_labels, plot_vals_on_bands


class Ephmat(CalcMode):
    """
    Class representation of a Perturbo ephmat calculation.

    Attributes
    ----------
    kpt : RecipPtDB
       Database for the k-points used in the ephmat calculation, containing N points.
    
    qpt : RecipPtDB
       Database for the q-points used in the ephmat calculation, containing M points.
    
    phdisp : UnitsDict
       Database for the phonon energies computed by the ephmat calculation. The keys are
       the phonon mode, and the values are an array (of length M) containing the energies at each q-point
       with units phdisp.units
    
    ephmat : UnitsDict
       Database for the e-ph matrix elements computed by the ephmat calculation. The keys are
       the phonon mode, and the values are an array (of length NxM) where element (n, m)
       is the e-ph matrix element (units ephmat.units) between an electron at k-point n and phonon at q-point m
    
    defpot : UnitsDict
       Database for the deformation potentials computed by the phdisp calculation. The keys are
       the phonon mode, and the values are an array (of length NxM) where element (n, m)
       is the deformation potential (units defpot.units) of an electron at k-point n and phonon at q-point m.

    """

    def __init__(self, pert_dict):
        """
        Constructor method

        Parameters
        ----------
        pert_dict : dict
            Dictionary containing the inputs and outputs from the ephmat calculation.

        """
        super().__init__(pert_dict)

        if self.calc_mode != 'ephmat':
            raise ValueError('Calculation mode for a Ephmat object should be "ephmat"')

        phdisp_units = self._pert_dict['ephmat'].pop('phonon energy units')
        defpot_units = self._pert_dict['ephmat'].pop('deformation potential units')
        ephmat_units = self._pert_dict['ephmat'].pop('e-ph matrix elements units')
        nmode = self._pert_dict['ephmat'].pop('number of phonon modes')

        kpath_units = self._pert_dict['ephmat'].pop('k-path coordinate units')
        kpath = np.array(self._pert_dict['ephmat'].pop('k-path coordinates'))
        kpoint_units = self._pert_dict['ephmat'].pop('k-point coordinate units')
        kpoint = np.array(self._pert_dict['ephmat'].pop('k-point coordinates'))
        
        qpath_units = self._pert_dict['ephmat'].pop('q-path coordinate units')
        qpath = np.array(self._pert_dict['ephmat'].pop('q-path coordinates'))
        qpoint_units = self._pert_dict['ephmat'].pop('q-point coordinate units')
        qpoint = np.array(self._pert_dict['ephmat'].pop('q-point coordinates'))

        ephmat_dat = self._pert_dict['ephmat'].pop('phonon mode')
        
        self.kpt = RecipPtDB.from_lattice(kpoint, kpoint_units, self.lat, self.recip_lat, kpath, kpath_units)
        self.qpt = RecipPtDB.from_lattice(qpoint, qpoint_units, self.lat, self.recip_lat, qpath, qpath_units)

        phdisp = {}
        defpot = {}
        ephmat = {}

        N = len(self.kpt.path)
        M = len(self.qpt.path)

        for phidx in ephmat_dat.keys():
            phdisp[phidx] = ephmat_dat[phidx].pop('phonon energy')
            defpot[phidx] = np.array(ephmat_dat[phidx].pop('deformation potential')).reshape(N, M)
            ephmat[phidx] = np.array(ephmat_dat[phidx].pop('e-ph matrix elements')).reshape(N, M)

        self.phdisp = UnitsDict.from_dict(phdisp, phdisp_units)
        self.defpot = UnitsDict.from_dict(defpot, defpot_units)
        self.ephmat = UnitsDict.from_dict(ephmat, ephmat_units)

    def plot_phdisp(self, ax, show_qpoint_labels=True, c='k', ls='-', energy_window=None):
        """
        Method to plot the phonon dispersion.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
           Axis on which to plot the phdisp.

        energy_window : tuple of int, optional
           The range of band energies to be shown on the y-axis.

        show_qpoint_labels : bool, optional
           If true, the q-point labels stored in the labels attribute will be shown on the plot. Default true.

        Returns
        -------
        ax: matplotlib.axes.Axes
           Axis with the plotted bands.

        """
        ax = plot_dispersion(ax, self.qpt.path, self.phdisp, self.phdisp.units, c, ls, energy_window)

        if show_qpoint_labels:
            ax = plot_recip_pt_labels(ax, self.qpt.labels, self.qpt.points, self.qpt.path)

        return ax

    def plot_defpot(self, ax, kpoint_idx=0, show_qpoint_labels=True, cmap='RdBu', energy_window=None):
        """
        Method to plot the phonon dispersion.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
           Axis on which to plot the phdisp.

        kpoint_idex : int, optional
            Index of the k-point to plot the deformation potentials for. Deformation potentials will be plotted along q-points, at this k-point
            By default, it will be the first k-point.

        energy_window : tuple of int, optional
           The range of band energies to be shown on the y-axis.

        show_qpoint_labels : bool, optional
           If true, the q-point labels stored in the labels attribute will be shown on the plot. Default true.

        Returns
        -------
        ax: matplotlib.axes.Axes
           Axis with the plotted bands.

        """

        values = {}

        for key, val in self.defpot.items():
            values[key] = self.defpot[key][kpoint_idx, :]

        ax = plot_vals_on_bands(ax, self.qpt.path, self.phdisp, self.phdisp.units, values=values, energy_window=energy_window, cmap=cmap)

        if show_qpoint_labels:
            ax = plot_recip_pt_labels(ax, self.qpt.labels, self.qpt.points, self.qpt.path)

        return ax

    def plot_ephmat(self, ax, kpoint_idx=0, show_qpoint_labels=True, cmap='RdBu', energy_window=None):
        """
        Method to plot the phonon dispersion.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
           Axis on which to plot the phdisp.

        kpoint_idex : int, optional
            Index of the k-point to plot the e-ph elements for. E-ph elements will be plotted along q-points, at this k-point
            By default, it will be the first k-point.
        energy_window : tuple of int, optional
           The range of band energies to be shown on the y-axis.

        show_qpoint_labels : bool, optional
           If true, the q-point labels stored in the labels attribute will be shown on the plot. Default true.

        Returns
        -------
        ax: matplotlib.axes.Axes
           Axis with the plotted bands.

        """

        values = {}

        for key, val in self.ephmat.items():
            values[key] = self.ephmat[key][kpoint_idx, :]

        ax = plot_vals_on_bands(ax, self.qpt.path, self.phdisp, self.phdisp.units, values=values, energy_window=energy_window, cmap=cmap)

        if show_qpoint_labels:
            ax = plot_recip_pt_labels(ax, self.qpt.labels, self.qpt.points, self.qpt.path)

        return ax
